// Adapted from extant sizeup code

var request = require('request');

var util = require('./util');

module.exports = function makeGetData(apiKey) {

  // Factory function (encapsulating an authentication context per apiKey) that
  // returns a function that gets data from the SizeUp data API via JSONP, authenticating as needed


  var me = {};
  me.sessionId  = null;
  me.apiToken   = null;
  me.instanceId = null;
  me.widgetToken = '';
  me.apiKey     = apiKey;
  me.callbackComplete = {};
  
  me.currentLocation = "https://api.sizeup.com";
  // The browser client computes this value by looking at where the script
  // was loaded from. This client doesn't have that, so we hardcode for now.
  // TODO make this config when we merge the two clients.

  var getData = function (path, params, onSuccess, onError) {
    var opts = { aborted: false };
    function abort() {
      opts.aborted = true;  // NOTE: why not null the success and error functions so they may be GC'd?
    }

    var r;
    if (!onSuccess && !onError) {
      r = new Promise(function(resolve, reject) {
        onSuccess = resolve;
        onError   = reject;
      })
      r.abort = abort;
    } else {
      r = { abort:abort };    // historical
    }

    // If auth is missing or old, and not already underway ...
    if ((!me.sessionId || me.authdAt && me.authdAt + 5*60*1000 < +new Date()) && !me.getDataQueue) {
      // ... Authenticate (async); Queue requests until auth is available.
      //  Alternately client could use a callback for auth.  Burden belongs here.
      //  TODO: if we could compute auth values from the apiKey client side, none of this would be needed; but I think server must establish session.
      //  TODO: Why doesn't server just accept key on each request? How are sessions used/implemented on server?

      me.getDataQueue = [];
      function applyToAndDeleteGetDataQueue(fn) {
        var q = me.getDataQueue;
        delete me.getDataQueue;  // delete before fn is called: must be missing or authentication is underway.

        for (var i = 0; i < q.length; i++) {                                    // TODO: move to es6??
          fn(q[i]);
        }
      }
      authenticate()
        .then(function () {
          applyToAndDeleteGetDataQueue(function (x) {                           // TODO: move to es6??
            getData(x.path, x.params, x.onSuccess, x.onError);
          });
        })
        .catch(function (message) {
          applyToAndDeleteGetDataQueue(function (x) {                           // TODO: move to es6??
            x.onError(message);
          });
        })
    }
    if (me.getDataQueue) {
      me.getDataQueue.push({
        path:       path,
        params:     params,
        onSuccess:  onSuccess,
        onError:    onError
      });
      return r;
    }

    var url = me.currentLocation + path;
    params['cb'] = 'JSONP_WRAPPER';
    var tokenUrl = buildTokenUrl(url, params);
    // NOTE: E.g., http://a2-api.sizeup.com/data/place/search/?term=fresno&maxResults=10&cb=sizeup.api.cbb.cb33&o=sizeup.com&s=1f4uhh94x0968t1x46oox9z3j&t=utZOqvvO8s2KUngekSFKXGCs8Xxxh9jIHzZcuNNyuROLRHA4MFBr%2BiqIWuk4Z39E&i=re6ktch2yfdd3wb3xocdi8zrh
    // NOTE: E.g., http://a2-api.sizeup.com/data/place/search/?term=fresno&maxResults=10
    //   &cb=CB
    //   &o=   document.location.hostname
    //   &s=   me.sessionId
    //   &t=   me.apiToken
    //   &i=   me.instanceId

    request(
     {url:tokenUrl, headers: {"User-Agent":"Mozilla/5.0"}}, // workaround for TECH-433
     function (error, response, body) {
      if (error || !response || response.statusCode!==200) {
        return onError(
          error    ? "Network error: " + error :
          response ? "Query failed (bad query?): "+response.statusMessage+"\n"+response.body :
          'unknown'
        );
      }

      try {
        // console.log(body);
        var unJsonpWrapped = /^JSONP_WRAPPER\((.*)\)$/.exec(body)[1];
        var result = unJsonpWrapped ? JSON.parse(unJsonpWrapped) : null;
      } catch (e) {
        return onError(e);
      }

      onSuccess(result);
    });

    return r;
  };


  var authenticate = function () {
    var authUrl = me.currentLocation + '/js/authenticate?key=' + apiKey;    
console.log(authUrl);
    return new Promise(function (resolve, reject) {
      request(
       {url:authUrl, headers: {"User-Agent":"Mozilla/5.0"}}, // workaround for TECH-433
       function (error, response, body) {
        if (!error && (response||{}).statusCode == 200) {
          me.authdAt = +new Date();

          var auth = JSON.parse(body);
          me.apiToken = auth.Token;
          me.sessionId = auth.SessionId;
          me.instanceId = auth.InstanceId;
          if (me.apiToken && me.sessionId && me.instanceId) {
            resolve();   // TODO: might be clearer to resolve with the auth'd context as argument... or not
          } else {
            reject("No auth tokens found at authentication URL " + authUrl);
          }

        } else {
          reject("Error attempting to authenticate: " + (
              error
              ? "Failed to contact sizeup server: " + error
              : "Auth error (invalid key?): "
              + (response ? ("status code: "+response.statusCode+"; "+response.statusMessage) : "unknown") ) );
        }
      })
    });
  };


  var buildTokenUrl = function (src, params){
    if (src.indexOf('?') < 0) {
      src = src + '?';
    }
    else {
      src = src + '&';
    }
    params['o'] = 'sizeup.com';  // TODO: unused??
    params['s'] = me.sessionId;
    params['t'] = me.apiToken;
    params['i'] = me.instanceId;
    if(me.widgetToken != '') {
      params['wt'] = me.widgetToken;
    }
    src = src + util.formatParams(params);
    return src;
  };


  return {
    authenticate: authenticate,
    buildTokenUrl: buildTokenUrl,
    getData: getData,
    me: me
  };
}
